#include "sys:/TextResponse.ms"

// Return this version # in the response headers so the Publisher API can tell if an incorrect script is being used
var ThisScriptVersion = "0.91"


function FileTypeFromExtension()
{
	var ext = req.getParameter("ext");
	if (ext.charAt(0) != '.')
		ext = '.' + ext;
	var mtype = Media.getTypeFromExtension(ext);
	logMsg("FileTypeFromExtension-ext: '" + ext + "', mtype: '" + mtype + "'");
	SetTextResponse(mtype);
}



// Main ingestion function. At a minimum, generates basic metadata for the specified doc. Can also generate a proxy tiff, PDF/Office links, and a thumb of a specified page.
function Ingest()
{
	var startTime = new Date().getTime();

	try
	{
		var workDir = req.getJobTempDir();
		var workThumb = null;
		var docFile = null;
		var params = {};
		params.useCMYK = false;

		var srcDocUrl = req.getParameter("srcdoc");
		var dstProxyName = req.getParameter("proxy");
		var dstMiscDataName = req.getParameter("miscdata");
		var dstThumbName = req.getParameter("thumb");
		var dstMetaName = req.getParameter("meta");
		params.dpi = req.getParameter("dpi") * 1;
		var page = req.getParameter("page") * 1;
		var thumbWidth = req.getParameter("thumbWidth") * 1;
		var thumbHeight = req.getParameter("thumbHeight") * 1;
		var bg = req.getParameter("bg") * 1;
		var maxPadding = req.getParameter("maxPadding") * 1;
		var cropPadding = !!req.getParameter("cropPadding");
		var checkerFg = req.getParameter("fg") || 0;
		var checkerBoxSize = req.getParameter("cbsz") || 0;

		resp.setHeader("X-CPUCoinMiner-PublisherScriptVer", ThisScriptVersion);

		logMsg("srcDocUrl: " + srcDocUrl + ", dstProxyName: " + dstProxyName + ", dstMiscDataName: " + dstMiscDataName + ", dstThumbName: " + dstThumbName);

		if (srcDocUrl && srcDocUrl.substr(0, 4) !== "http")
			throw "Srcdoc must use the HTTP protocol. Operation aborted.";

		// Copy the document from the dApp into the temp workdir so we always have a filename w/ an extension to use
		var ep = srcDocUrl.lastIndexOf(".");		// Publisher API app guarantees last queried filename is the file we're interested in, so use this to get its extension
		if (ep < 0)
			throw "Srcdoc doesn't have a filename extension thus cannot be ingested: " + srcDocUrl;
		params.name = workDir + "/srcdoc"+ srcDocUrl.substring(ep).toLowerCase();
		srcDocUrl += (srcDocUrl.lastIndexOf("?") < 0 ? "?_=" : "&_=") + req.getJobId();
		docFile = new File(srcDocUrl);
		docFile.copy(params.name);
		docFile.clearCached();			// clear the downloaded original doc from the fsnet cache so we don't pollute the miner (the temp copy gets deleted automatically from the jobtemp folder when we exit)
		docFile = null;

		var xferTime = new Date().getTime();

		var m = new Media();
		var fmt = getDocType(params.name);

		// Prepare a zip object to return all of the files requested
		var respPayload = new Zip();

		// For Office docs, generate a PDF file. Subsequent processing will work with that PDF file rather than the original document.
		// NOTE: Unlike the current SPC implementation, these PDFs are converted to multipage tif for unizoom and the PDF is thrown away at the end of ingest.
		if (fmt.isOffice)
		{
			if (fmt.isExcel)
			{
				params.imageHeight = 1000;
				params.imageWidth = 1000;
			}

			// Always generate PDF for office docs because at the very least we need to render it to get width, height, and pagecount
			params.intermediateFilename = workDir + "/officedoc.pdf";
			logMsg("Generating PDF file for Office doc (" + params.intermediateFilename + ")");
			m.load(params);		// doesn't load anything, just converts office doc to PDF in {jobtempdir}/officedoc.pdf"
			logMsg("PDF generated.");
			params.name = params.intermediateFilename;
			delete params.intermediateFilename;			// need to remove this field otherwise it can confuse loading further down into generating another intermediate file
			fmt.isPdf = true;							// now that the office doc is a pdf, treat it like one so the rest of the info can be acquired
		}

		// Handle links extraction from PDF and Office docs if requested
		if (fmt.isPdf && dstMiscDataName)
		{
			var lf, linksPath = workDir + "/links.json";

			if (linksPath)
			{
				logMsg("Generating " + linksPath + " from " + params.name + " @ " + params.dpi + " dpi");
				try
				{
					lf = new File(linksPath);
					lf.write(new TextExtraction(params.name).getLinks("json", params.dpi, 0, 0));
					lf.close();
					respPayload.addFile(linksPath, dstMiscDataName);		// return this file as part of the response payload
				}
				catch (ex)
				{
					logMsg("Links generation FAILED - " + ex, LogFile.SeverityError);
				}
			}
		}

		var width = -1;
		var height = -1;
		var pageCount = -1;

		// Always extract at least the first page of these formats to a TIFF so we can get accurate width and height values, even if we don't send the TIFF back to the dApp
		if (fmt.isPdf || fmt.isDcRaw || fmt.isFlash)	// don't need to check for office type because ghostscript is set if dstProxyName is present
		{
			var workTiffPath = workDir + "/proxy.tif";

			try
			{
				// For vector and office formats, create a huge multipage TIFF for unizoom, but only if we're asked to send this TIFF back to the dApp
				if (dstProxyName && fmt.isPdf)
				{
					logMsg("Making multipage TIFF " + workTiffPath + " at " + params.dpi + "dpi");
					params.intermediateFilename = workTiffPath;
					m.load(params); 							// this converts the PDF to a TIFF using VI's SDK - it never even goes into the Media object. Also the TIFF is forced to RGB
					delete params.intermediateFilename;			// need to remove this field otherwise it can confuse loading further down into generating another intermediate file
				}
				else
				{
					// If no multipage TIFF needed, just make the first page into a TIFF for speed, basic metadata and optionally, a thumbnail
					// However, if we are not returning the TIFF with PDFs, converted office docs, raw, and Flash, we still need to get a page count from a non-rasterized document since we're only going to be rendering page 1.
					if (!dstProxyName && (fmt.isPdf || fmt.isDcRaw || fmt.isFlash))
					{
						logMsg("Getting page count");
						var info = Media.getFileInfo(name @ params.name);
						pageCount = info.Frames || 1;
						logMsg("Got page count: "+pageCount);
					}
					logMsg("Making single page TIFF " + workTiffPath + " at " + params.dpi + "dpi");
					params.dontCache = true;
					params.applyEmbedded = true;
					if (page)
						params.page = page * 1;		// if get_page is specifying a page when no single page proxy is present, make sure we tell the reader to render the specified one
					else
						params.page = 1;			// otherwise just grab the first page
					params.frame = params.page;		// need to do this for gif and tiff
					m.loadAsRgb(params);
					width = m.getWidth();
					height = m.getHeight();
					m.save(name @ workTiffPath);

					// As a optimization, since we already have the raster in memory don't bother loading the TIFF again below to make the thumb.
					if (dstThumbName)
					{
						try
						{
							m = scaleImage(m, thumbWidth, thumbHeight, bg, "Smooth", maxPadding, cropPadding, checkerFg, checkerBoxSize);
							workThumb = workDir + "/thumb.jpg";
							m.save(name @ workThumb);
						}
						catch (ex)
						{
							dstThumbName = null;
							logMsg("Thumbnail conversion FAILED - " + ex, LogFile.SeverityError);
						}
					}
				}
				params.name = workTiffPath;	// we'll work with this file from now on instead of the original or office proxy PDF
			}
			catch (ex)
			{
				dstProxyName = null;
				logMsg("Proxy TIFF generation FAILED - " + ex, LogFile.SeverityError);
			}

			// Only send the TIFF back to the dApp if requested and succeeded
			if (dstProxyName)
				respPayload.addFile(workTiffPath, dstProxyName);
		}

		// Generate thumbnail for page 1 of document, if requested
		if (dstThumbName && !workThumb)
		{
			try
			{
				if (page)
					params.page = page * 1;		// if get_page is specifying a page when no single page proxy is present, make sure we tell the reader to render the specified one
				else
					params.page = 1;			// otherwise just grab the first page
				params.frame = params.page;		// need to do this for gif and tiff
				logMsg("getting page " + params.frame);
				m.load(params);
				m = scaleImage(m, thumbWidth, thumbHeight, bg, "Smooth", maxPadding, cropPadding, checkerFg, checkerBoxSize);
				workThumb = workDir + "/thumb.jpg";
				m.save(name @ workThumb);
			}
			catch (ex)
			{
				dstThumbName = null;
				logMsg("Thumbnail conversion FAILED - " + ex, LogFile.SeverityError);
			}
		}

		// If we generated a thumb, make it part of the payload
		if (dstThumbName)
			respPayload.addFile(workThumb, dstThumbName);

		// Get basic metadata for this document, if we don't have it already.
		if (dstMetaName)
		{
			if (width < 0 || height < 0 || pageCount < 0)
			{
				var info = {};
				try
				{
					info = Media.getFileInfo(name @ params.name);
				}
				catch (ex)
				{
					logMsg("Basic metadata extract FAILED for '" + params.name  + "' - " + ex, LogFile.SeverityError);
				}
				if (width < 0 && info.Width)
					width = info.Width;
				if (height < 0 && info.Height)
					height = info.Height;
				if (pageCount < 0)
					pageCount = info.Frames || 1;
			}

			//var metadata = '{ "fileType": "JPEG", "pageCount": 1, "width": 640, "height": 480, "foo": 1 }';
			var metadata = '{ "fileType": "' + fmt.fileType + '", "pageCount": ' + pageCount + ', "width": ' + width + ', "height": ' + height + ' }';
			logMsg("Basic metadata: " + metadata);
			var mf = new File(workDir + "/metadata.json");
			mf.write(metadata);
			mf.close();
			respPayload.addFile(mf.getFilePath(), dstMetaName);
		}

		// Generate the payload file and stream it back w/ no caching
		resp.setObject(respPayload, RespType.Streamed);
	}
	catch (e)
	{
		if (docFile)
			docFile.clearCached();
		resp.setStatusCode(500); // DMP: commented out status code setting because it causes the dApp web req to throw w/o being able to see the error in the header set below (uncommented since CGN needs 500 to reissue work to another miner)
		logMsg("Ingest() exception: " + e, LogFile.SeverityError);
		resp.setHeader("X-CPUCoinMiner-Error", makeAttributeValue(e.toString()))
	}

	var stopTime = new Date().getTime();
	var timing = (xferTime - startTime) + "/" + (stopTime - xferTime) + "/" + (stopTime - startTime);
	resp.setHeader("X-CPUCoinMiner-Timing", timing);
	logMsg("Ingest() " + timing + " ms for xfer/processing/total of " + srcDocUrl + " to " + fmt);
}



// Add a thumbnail to an existing gallery, or add a thumbnail to a blank canvas if no existing gallery is specified and upload the new gallery image to the remote dApp server.
function BuildGallery()
{
	var startTime = new Date().getTime();

	try
	{
		var srcCanvasUrl = req.getParameter("srccanvas");
		var brushUrl = req.getParameter("brush");
		var cols = req.getParameter("cols") * 1
		var rows = req.getParameter("rows") * 1
		var bg = req.getParameter("bg") * 1
		var cw = req.getParameter("w") * 1;
		var ch = req.getParameter("h") * 1;
		var bw = req.getParameter("tw") * 1;
		var bh = req.getParameter("th") * 1;
		var x = req.getParameter("x") * 1;
		var y = req.getParameter("y") * 1;

		resp.setHeader("X-CPUCoinMiner-PublisherScriptVer", ThisScriptVersion);

		var type;
		var canvas = new Media();
		if (srcCanvasUrl)
		{
			type = getMediaType(srcCanvasUrl);
			if (!type)
			{
				logMsg("BuildGallery: Unsupported filetype encountered - " + srcCanvasUrl, LogFile.SeverityWarning);
				return;
			}
			srcCanvasUrl += (srcCanvasUrl.lastIndexOf("?") < 0 ? "?_=" : "&_=") + req.getJobId();
			canvas.load({name: srcCanvasUrl, type: type});
			new File(srcCanvasUrl).clearCached();	//	don't leave these around - it causes the updated version to not get used, plus we don't want to leave stuff on the miner, anyway
		}
		else
			canvas.makeCanvas({xs: cw, ys: ch, fillColor: bg});

		var brush = new Media();
		if (brushUrl)
		{
			type = getMediaType(brushUrl);
			if (type)
			{
				brushUrl += (brushUrl.lastIndexOf("?") < 0 ? "?_=" : "&_=") + req.getJobId();
				brush.loadAsRgb({name: brushUrl, type: type, dontCache: true, applyEmbedded: true});
				new File(brushUrl).clearCached();
				brush.scale({xs: bw, ys: bh, constrain: true, padColor: bg, alg: "best"});
				canvas.composite({source: brush, x: x, y: y});
			}
		}

		for (var c = 1; c < cols; c++)
			canvas.line({x1: c*bw, x2: c*bw, y1: 0, y2: ch, color: 0xaaaaaa});
		for (var r = 1; r < rows; r++)
			canvas.line({x1: 0, x2: cw, y1: r*bh, y2: r*bh, color: 0xbbbbbb});

		// Return image as JPEG in response; skip all caching
		resp.setObject(canvas, RespType.Streamed, {type: "JPEG"});
	}
	catch (e)
	{
		//resp.setStatusCode(500);
		logMsg("BuildGallery() exception: " + e, LogFile.SeverityError);
		resp.setHeader("X-CPUCoinMiner-Error", makeAttributeValue(e.toString()))
	}

	logMsg((new Date().getTime() - startTime) + " ms for gallery of " + (brushUrl?brushUrl:"canvas only"));
}



// Render a zoomed image based on level and xy coords. Scale the image to a specified w & h. Allow optional checkboard backdrop for alpha images.
function Zoom()
{
	var startTime = new Date().getTime();

	// Constants
	var JpegQuality = 95;
	var CheckerBoxSize = 50;
	var ScaleAlgorithm = "best"; // os, best, smooth, fast

	try
	{
		var srcUrl = req.getParameter("srcurl");
		var level = req.getParameter("level") * 1;
		var xcoord = req.getParameter("x") * 1;
		var ycoord = req.getParameter("y") * 1;
		var canvasW = req.getParameter("outW") * 1;
		var canvasH = req.getParameter("outH") * 1;
		var page = req.getParameter("page") * 1;
		var dpi = req.getParameter("dpi") * 1;
		var bgColor = req.getParameter("bgColor") || 0;
		var checkerFg = req.getParameter("checkerColor") || 0;
		var checkerBoxSize = req.getParameter("checkerBoxSize") || 0;

		resp.setHeader("X-CPUCoinMiner-PublisherScriptVer", ThisScriptVersion);

		var image = new Media();
		var type = getMediaType(srcUrl);
		if (type)
		{
			srcUrl += (srcUrl.lastIndexOf("?") < 0 ? "?_=" : "&_=") + req.getJobId();
			image.loadAsRgb({name: srcUrl, type: type, frame: page, page: page, dpi: dpi, usecmyk: false, dontCache: true, applyEmbedded: true});
			new File(srcUrl).clearCached();

			var xferTime = new Date().getTime();

			/*if (fileIsSchawkTIFF(srcUrl))
				image.convert(rtype @ "RGB-24");
			else*/ if (image.getImageFormat() == "32 Bit RGBA" && checkerBoxSize && checkerFg != bgColor)	// to make alpha channel not black, we have to composite the image on desired-colored background if caller wants to do that
				image = compositeOnCheckerboard(image, checkerFg, bgColor, checkerBoxSize);
			image.setColor(backColor @ bgColor);
			image.zoom(xs @ canvasW, ys @ canvasH, scale @ level, x @ xcoord, y @ ycoord, alg @ ScaleAlgorithm);

			var canvas = new Media();
			canvas.makeCanvas(xs @ canvasW, ys @ canvasH, fillColor @ bgColor);
			canvas.composite(source @ image, x @ canvasW / 2, y @ canvasH / 2);

			var stopTime = new Date().getTime();
			var timing = (xferTime - startTime) + "/" + (stopTime - xferTime) + "/" + (stopTime - startTime);
			resp.setHeader("X-CPUCoinMiner-Timing", timing);
			logMsg("Zoom() " + timing + " ms for xfer/processing/total of " + srcUrl);

			resp.setObject(canvas, RespType.Streamed, {type: "JPEG", quality: JpegQuality});		// stream back as a JPEG w/ the specified quality (no caching)
		}
		else
			logMsg("Warning - document passed to Zoom() isn't supported: " + srcUrl, LogFile.SeverityWarning);
	}
	catch (e)
	{
		resp.setStatusCode(500);
		logMsg("Zoom() exception: " + e, LogFile.SeverityError);
		resp.setHeader("X-CPUCoinMiner-Error", makeAttributeValue(e.toString()))
	}
}



function PDFExport()
{
	var srcPath;
	srcPath = req.getParameter("path");
	if (!srcPath)
		throw "No file for export.";

	var srcName = srcPath.substr(srcPath.lastIndexOf("&path=") + 6);
	var dstName = makeAttributeValue(new File(srcName).getFileNameNoExt() + ".pdf");	// use this to save the file to the user's system w/ a friendly filename

	var tenantId = req.getParameter("tenantId");
	var siteId = req.getParameter("siteId");
	var listId = req.getParameter("listId");
	var itemId = req.getParameter("itemId");
	var rev = req.getParameter("revision");
	var hideAnnotations = req.getParameter("hideAnnotations");
	var redactMode = req.getParameter("redactMode");
	var redactOpacity = req.getParameter("redactOpacity");
	var workDir = req.getJobTempDir() + "/";

	resp.setHeader("X-CPUCoinMiner-PublisherScriptVer", ThisScriptVersion);

	if (hideAnnotations && (hideAnnotations === "1" || hideAnnotations === "yes"))
		hideAnnotations = true;
	else
		hideAnnotations = false;
	if (!redactMode)
		redactMode = "off";
	if (redactOpacity)
	{
		redactOpacity = Math.round(redactOpacity * 255);	// convert 0.0 through 1.0 to 0 through 255
		if (redactOpacity > 255)							// make sure values fall into a valid range
			redactOpacity = 255;
		else if (redactOpacity < 0)
			redactOpacity = 0;
	}
	else
		redactOpacity = 255;

	// Error on mutually exclusive param values
	if (hideAnnotations && redactMode === "attributed")
		throw 'hideAnnotations cannot be enabled when redactMode is set to "attributed".';

	var m = new Media();
	var om = new Media();	// this is our output media. We add pages to it as needed.
	var nm = new Media();	// a work media object for misc rendering
	var note, i;

	var commentWidth = 240;	// width of the comments column in pixels
	var indent = 4;			// # of pixels to indent the comment bar from the right side of the image
	var dshadSize = 5;		// size of the sidebar comments dropshadow in pixels

	var typeInfo = getDocType(srcPath);
	if (typeInfo.isOffice)
	{
		// For Office docs, first convert them into a PDF and make the PDF the new source file.
		m.load({name: srcPath, intermediatefilename: workDir + "inter.pdf" /*, type: typeInfo.fileType*/ });
		srcPath = workDir + "inter.pdf";
		typeInfo.isPdf = true;	// fall thru to code below
		//typeInfo.fileType = "PDF";
	}
	if (typeInfo.isPdf)
	{
		// For VI-handled types, convert them info TIFFs so we can page through them quickly w/o having to load them entirely into memory
		m.load({name: srcPath, intermediatefilename: workDir + "inter.tif", dpi: 150 /*, type: typeInfo.fileType*/ });
		srcPath = workDir + "inter.tif";
		//typeInfo.fileType = "TIFF";
	}
	/*else
	{
		// For all other filetypes, we need to copy the file into the workdir otherwise getFileInfo() will fail
		new File(srcPath).copy(workDir + srcName);
		srcPath = workDir + srcName;	// srcName has the filename extension so getFileInfo() will be happy
	}*/

	// Main page loop
	var info = Media.getFileInfo(name @ srcPath);
	if (!info.Frames || (!typeInfo.isOffice && !typeInfo.isPdf))
		info.Frames = 1;

	// Load all of the frames and figure out the largest dimensions. We will use that at the w&h for the entire doc
	var mw = 0, mh = 0;
	for (var p=1; p<=info.Frames; p++)
	{
		m.load(name @ srcPath, frame @ p);
		var iw = m.getWidth(), ih = m.getHeight();
		if (mw < iw)
			mw = iw;
		if (mh < ih)
			mh = ih;
	}

	//logMsg("PAGES="+info.Frames+", SRCPATH="+srcPath+", DSTNAME="+dstName, LogFile.SeverityAlways);
	// Loop through all pages
	for (p=1; p<=info.Frames; p++)
	{
		var notes = [];

		// Sort the array in ascending order by last modified time
		notes.sort(timeCmp);

		// Load just the page we are working on from either the original file (if it's a raster) or the intermediate TIFF (if it's a vector or doc format)
		m.load(name @ srcPath, frame @ p);

		// Force image of page to 24-bit RGB since we'll be saving it as a JPEG-compressed PDF
		var notCorrected = true;
		if (m.getPixelFormat().substr(0, 4) == "CMYK")
		{
			m.colorCorrect(sourceprofile @ "cmyk", destprofile @ "RGB", overrideembedded @ false);
			notCorrected = false;
		}
		if (m.getPixelFormat().indexOf("A") >= 0)
		{
			m.composite(color @ 0xFFFFFF, blend @ "Under");		// create white backdrop for alpha channeled images before stripping the alpha
			m.convert(rtype @ "RGB");							// otherwise the backdrop uses whatever is in the color channels which can look strange
		}
		else if (m.getPixelFormat() !== "RGB-24")
			m.convert(rtype @ "RGB");
		if (notCorrected && m.embeddedProfile())
			m.colorCorrect(destprofile @ "RGB", overrideembedded @ false);	
		//m.scale(xs @ 320, constrain @ true, alg @ "Best");
		iw = m.getWidth(), ih = m.getHeight();

		// Now we're ready to render the selection boxes onto the loaded page
		var smallestDim = iw < ih ? iw : ih;
		var lw = Math.ceil(smallestDim / 1000);		// take smallest dimension and increase line width by 1 for every 1000 pixels of image size
		for (i=0; i<notes.length; i++)
		{
			try
			{
				note = notes[i];
				var isRedaction = note.quote === "__R~~";

				// Skip rendering selection if this is a sidebar-only comment which has no selection box to render
				if (!note.image_selection || !note.image_selection.w || !note.image_selection.h)
					continue;

				if (isRedaction)
				{
					// Skip rendering selection if this is a redaction and redactions are disabled
					if (redactMode === "off")
						continue;
				}
				else if (hideAnnotations)	// at this point we know it must be a highlighted annotation, so skip it if rendering of annotations is turned off
					continue;

				// Render the box for both annotations and redactions
				var x = parseFloat(note.image_selection.x.slice(0, -1)) / 100.0;
				var y = parseFloat(note.image_selection.y.slice(0, -1)) / 100.0;
				var w = parseFloat(note.image_selection.w.slice(0, -1)) / 100.0;
				var h = parseFloat(note.image_selection.h.slice(0, -1)) / 100.0;
				var c = isRedaction ? "#000" : note.user.color;
				c = parseInt(c.substr(1,1) + c.substr(1,1) + c.substr(2,1) + c.substr(2,1) + c.substr(3,1) + c.substr(3,1), 16);
				x *= iw;
				y *= ih;
				w *= iw;
				h *= ih;

				m.rectangle(color @ c, x @ x, y @ y, xs @ w, ys @ h, opacity @ isRedaction ? redactOpacity : 77, fill @ true);
				if (lw != 0)		// skip drawing the outline around the selection box and number if the image is ridiculously tiny
				{
					if (!isRedaction)		// don't draw outlines for redactions
					{
						var xs = w-lw-lw;
						var ys = h-lw-lw;
						if (xs < 1)
							xs = lw;
						if (ys < 1)
							ys = lw;
						m.rectangle(color @ 0, x @ x, y @ y, xs @ w, ys @ h, opacity @ 77, fill @ false, width @ lw);
						m.rectangle(color @ 0xFFFFFF, x @ x+lw, y @ y+lw, xs @ xs, ys @ ys, opacity @ 77, fill @ false, width @ lw);
					}
					if (!isRedaction || redactMode === "attributed")	// only draw the number if this isn't a redaction or user wants the redactions to be attributed
					{
						var nsize = Math.ceil(smallestDim / 50);
						if (nsize < 3)
							nsize = 3;
						nm.makeText(text @ (i+1)+".", justify @ "left", cleartype @ true, font @ "Arial", size @ nsize, style @ "bold", kern @ true, color @ 0xFFFFFF);
						nm.glow(opacity @ 255, color @ 0x000000, size @ 1, resizeCanvas @ true);
						var yy = y + 3, xx = x + 5;
						if (xx < 0)  xx = 0;
						else if (xx + nsize + 2 > iw)  xx = iw - nsize - 2;
						if (yy < 0)  yy = 0;
						else if (yy + nsize + 2 > ih)  yy = ih - nsize - 2;
						m.composite(source @ nm, handlex @ "left", handley @ "top", x @ xx, y @ yy);
					}
				}
			}
			catch (e) {
				logMsg("Failed to render annotation selection box: "+e, LogFile.SeverityError);
			}
		}

		// If pages are multiple sizes, we must make every one the single, largest size for positioning the sidebar
		if (iw != mw || ih != mh)
		{
			nm.makeCanvas(rtype @ "RGB-24", xs @ mw, ys @ mh, fillcolor @ 0xFFFFFF);
			m.composite(source @ nm, onto @ true);
		}

		// Skip all of the sidebar code if we're not showing annotations (redactions are only allowed to be rendered unattributed in this mode so there's no need to worry about that)
		if (!hideAnnotations)
		{
			// Create the sidebar for the comments
			m.crop(xo @ 0, xs @ mw+commentWidth+dshadSize+1+indent, padcolor @ 0xFFFFFF, relative @ false);

			// Now render the actual comments on the right side of the page, stopping before we go past the bottom
			y = 0;		// keep tabs on where we are vertically
			var t = new Media();
			for (i=0; i<notes.length; i++)
			{
				try
				{
					note = notes[i];
					if (note.quote !== "__R~~" || redactMode === "attributed")		// skip displaying users and their comments for redactions unless caller wants redactions to be attributed
					{
						c = note.user.color;
						c = parseInt(c.substr(1,1) + c.substr(1,1) + c.substr(2,1) + c.substr(2,1) + c.substr(3,1) + c.substr(3,1), 16);
						var tinfo1 = m.sizeText(text @ note.text, justify @ "left", font @ "Segoe UI", size @ 14, cleartype @ true, kern @ true, wrap @ commentWidth-22);
						var tinfo2 = m.sizeText(text @ "|", justify @ "left", font @ "Segoe UI", size @ 14, cleartype @ true, kern @ true);
						t.makeCanvas(rtype @ "RGB-24", xs @ commentWidth, ys @ tinfo1.height + tinfo2.height*2, fillcolor @ 0xFFFFFF);	// combine both the height of the comment and the height of a single line *2 to make room for the time and username
						t.rectangle(x @ 0, y @ 0, xs @ 20, ys @ t.getHeight(), color @ c, fill @ true, opacity @ 160);
						nm.makeText(text @ i+1, color @ 0xFFFFFF,  size @ 24, font @ "Arial", style @ "bold", kern @ true, cleartype @ true);
						if (nm.getWidth() > 20)
							nm.scale(xs @ 20, alg @ "best", constrain @ false);
						t.composite(source @ nm, x @ 10);
						t.drawText(text @ note.text, x @ 21, y @ 0, justify @ "left", font @ "Segoe UI", size @ 14, cleartype @ true, kern @ true, wrap @ commentWidth-22, color @ 0, handlex @ "left", handley @ "top");
						t.drawText(text @ "Created "+friendlyTime(note.updated)+"\n"+note.user.title, x @ 21, y @ t.getHeight(), justify @ "left", font @ "Calibri", size @ 10, cleartype @ true, kern @ true, style @ "bold", color @ 0x222222, handlex @ "left", handley @ "bottom");
						t.dropShadow(opacity @ 200, blur @ 4, dx @ dshadSize, dy @ dshadSize, color @ 0, resizeCanvas @ true);

						// If this comment is below the bottom of the image, we must insert a new duplicate page to handle the overflow
						var ny = y + t.getHeight();
						if (ny > mh && y != 0)
						{
							om.frameAdd(source @ m);																	// output the frame we're working on - we're done with it
							m.rectangle(x @ mw+indent, y @ 0, xs @ commentWidth+dshadSize+1, ys @ m.getHeight(), color @ 0xFFFFFF, fill @ true);	// clear the comments sidebar
							y = 0;																						// since we're starting from the top again
							ny = t.getHeight()
						}
						m.composite(source @ t, x @ mw+indent, y @ y, handlex @ "left", handley @ "top");
						y = ny;
					}
				}
				catch (e)
				{
					logMsg("Failed to render annotation sidebar comment: "+e, LogFile.SeverityError);
				}
			}
		}

		// Add the frame we were working on to the output media
		om.frameAdd(source @ m);
	} // end page loop

	resp.setMimeType("application/pdf");
	resp.setHeader("Content-Disposition", "inline; filename=\"" + dstName + "\"");
	resp.setObject(om, RespType.Streamed, { type: "pdf", jpegcompression: true });
}



// Internal utility functions defined below ---------------------------------


// Scale an image, limiting to w, h, or both. Also allow limiting to amt. of padding that is visible, instead cropping the image to make it appear as such.
function scaleImage(m, targetW, targetH, bg, alg, maxPadding, cropPadding, checkerFg, checkerBoxSize)
{
	var iw = m.getWidth();
	var ih = m.getHeight();
	checkerBoxSize *= 1;	// deal with ScriptEase bug where 0 != "0"

	// If a max ratio is set but caller only specified 1 dimension, assume the other is meant to be the same as the one they specified.
	if (maxPadding < 100)
	{
		if (!targetW && !targetH)		// if caller specified a maxPadding but didn't set any constraints, assume they want a square image limited to the image's smaller dimension
		{
			if (iw > ih)
				targetW = targetH = ih
			else
				targetW = targetH = iw;
		}
		else if (!targetW)
			targetW = targetH;
		else if (!targetH)
			targetH = targetW;
		// else both are specified so we're already all set
	}

	if (targetW && targetH)	// the fully constrained case
	{
		var didScale = false;
		var percentW = targetW / iw;	// these are the % the new dimension will be adjusted bv. If >1, we're scaling up!
		var percentH = targetH / ih;
		var verticallyLimited = percentH < percentW;
		if (maxPadding < 100)
		{
			maxPadding /= 100;				// convert percentage to rational
			var padding;
			if (verticallyLimited)
			{
				// The top and bottom of the image hit our height limit before the width, so scale to height
				var calcW = Math.round(iw * percentH);					// calc the new width based on height ratio since we're limited by that
				padding = targetW - calcW;								// this is the amt of padding that was added to the sides
				if (padding / targetW > maxPadding)						// is the image more padding on the sides than allowed?
				{
					calcW = targetW - maxPadding * targetW;				// calc what width needs to be so padding == maxPadding
					m.scale({ xs: calcW, constrain: true, alg: alg });	// scale image to the not-quite-fitting size to satisfy the maxPadding requirement
					m.crop({ xs: cropPadding ? calcW : targetW, ys: targetH, padColor: bg, transparency: 255 });
					didScale = true;
				}
			}
			else
			{
				// The sides of the image hit our width before the height, so scale to width
				var calcH = Math.round(ih * percentW);					// calc the new height based on width ratio since we're limited by that
				padding = targetH - calcH;								// this is the amt of padding that was added to the sides
				if (padding / targetH > maxPadding)						// is the image more padding on the top and bottom than allowed?
				{
					calcH = targetH - maxPadding * targetH;				// calc what height needs to be so padding == maxPadding
					m.scale({ ys: calcH, constrain: true, alg: alg });	// scale image to the not-quite-fitting size to satisfy the maxPadding requirement
					m.crop({ xs: targetW, ys: cropPadding ? calcH : targetH, padColor: bg, transparency: 255 });
					didScale = true;
				}
			}
		}

		// If we just needed a plain double-constrained scale, do it here
		if (!didScale)
		{
			if (cropPadding)
			{
				// Don't let scale() pad the image - calc the w & h ourselves so we only have the image itself
				if (verticallyLimited)
					m.scale({ xs: Math.round(iw * percentH), ys: targetH, constrain: false, alg: alg });
				else
					m.scale({ xs: targetW, ys: Math.round(ih * percentW), constrain: false, alg: alg });
			}
			else
			{
				m.scale({ xs: targetW, ys: targetH, constrain: true, alg: alg, padColor: bg, transparency: 255 });
			}
		}
	}
	else if (targetW)		// if we get here, one of the two has to be unset, so check which one. This is for constrained by width
	{
		m.scale({ xs: targetW, constrain: true, alg: alg });
	}
	else if (targetH)		// and this is constrained only by height
	{
		m.scale({ ys: targetH, constrain: true, alg: alg });
	}
	// if targets are unset, don't do anything and just return image at original size

	// If this has an alpha, fill in the transparent areas with the specified bgcolor
	if (m.getPixelFormat().indexOf("A-") > 0)
	{
		if (checkerBoxSize && checkerFg != bg)
			m = compositeOnCheckerboard(m, checkerFg, bg, checkerBoxSize);
		else
			m.composite({ color: bg, blend: "Under" });
	}

	return m;
}



function compositeOnCheckerboard(image, fgColor, bgColor, checkerBoxSize)
{
	var width = image.getWidth();
	var height = image.getHeight();
	var canvas = new Media();
	canvas.makeCanvas(xs @ width, ys @ height, fillColor @ fgColor);

	// Prepare checkboard backdrop
	var cbImage = new Media();
	cbImage.makeCanvas(xs @ checkerBoxSize * 2, ys @ checkerBoxSize * 2, fillColor @ fgColor);
	cbImage.rectangle(x @ 0, y @ 0, xs @ checkerBoxSize, ys @ checkerBoxSize, color @ bgColor, fill @ true);
	cbImage.rectangle(x @ checkerBoxSize, y @ checkerBoxSize, xs @ checkerBoxSize, ys @ checkerBoxSize-1, color @ bgColor, fill @ true);
	canvas.composite(source @ cbImage, x @ 0, y @ 0, handleX @ "left", handleY @ "top", tile @ true);

	// Composite 32-bit page onto it
	canvas.composite(source @ image, x @ width / 2, y @ height / 2);
	return canvas;
}



function getMediaType(name)
{
	var mtype = "";
	var dotPos = name.lastIndexOf(".");		// Publisher API app guarantees last filename is the file we're interested in, so use this to get its extension
	if (dotPos >= 0)
	{
		var ext = name.substring(dotPos).toLowerCase();
		mtype = Media.getTypeFromExtension(ext);
	}
	else
		logMsg("Warning - no filename extension found in " + name, LogFile.SeverityWarning);

	return mtype;
}



function getDocType(imgName)
{
	var ro = {};

	if (imgName)
	{
		ro.fileType = getMediaType(imgName);
		switch (ro.fileType)
		{
			case "MSExcel": ro.isExcel = true;
			case "MSWord":
			case "PowerPoint":
			case "OODraw": ro.isOffice = true; break;

			case "EPS": // .eps, .epsf
			case "PDF": // .pdf
			case "Postscript": // .ps
			case "Illustrator": // .ai
				ro.isPdf = true; break;

			case "CameraRaw": ro.isDcRaw = true; break;
			case "SWF": ro.isFlash = true; break;
		}
	}

	return ro;
}



/*function fileIsSchawkTIFF(filePath) { // these files have multiple garbage alpha channels that have to be removed
	var file = new File(filePath);
	var ext = file.getFileExtension();
	var fileInfo = Media.getFileInfo(name @ filePath);
	return ext == ".tif" && fileInfo.Width == 3144 && fileInfo.Height == 3144 && fileInfo.Format == "RGBA-32" && fileInfo.XDpi == 300;
}*/


function logMsg(msg, severity) {
	var f = new LogFile("Publisher.log");
	f.appendLine(msg, severity || LogFile.SeverityInfo); // SeverityAlways SeverityWarning
}



function makeAttributeValue(str)
{
	var illegalChars = ";\"/\\\r\n:*?<>|";

	var result = "";
	for (var i = 0; i < str.length; ++i)
	{
		var ch = str.charAt(i);
		if ((str.charCodeAt(i) & 0xFF80) != 0)
			result += "_";
		else if (illegalChars.indexOf(ch) != -1)
			result += "_";
		else
			result += ch;
	}

	return result;
}
