#include "sys:TextResponse.ms"
//#include "sys:ExifMetadata.ms"

// Return this version # in the response headers so the TweekIt API can tell if an incorrect script is being used
var ThisScriptVersion = "0.13"

function GetImage()
{
	var startTime = new Date().getTime();

	try
	{
		var workDir = req.getJobTempDir();
		var docFile = null;
		var params = {};
		params.useCMYK = false;
		params.loadMetadata = true;

		var srcDocUrl = req.getParameter("srcdoc");
		params.dpi = (req.getParameter("dpi") || 0) * 1;
		var outDpi = (req.getParameter("outDpi") || 72) * 1;
		var page = (req.getParameter("pg") || 0) * 1;
		var w = (req.getParameter("w") || 0) * 1;
		var h = (req.getParameter("h") || 0) * 1;
		var scale = (req.getParameter("scale") || 0) * 1;
		var x1 = (req.getParameter("x1") || 0) * 1;
		var y1 = (req.getParameter("y1") || 0) * 1;
		var x2 = (req.getParameter("x2") || 0) * 1;
		var y2 = (req.getParameter("y2") || 0) * 1;
		var bg = (req.getParameter("bg") || 0) * 1;
		var maxPadding = (req.getParameter("maxPadding") || 100) * 1;
		var cropPadding = !!((req.getParameter("cropPadding") || 0) * 1);
		var getInfo = !!((req.getParameter("gi") || 0) * 1);
		var elliptical = !!((req.getParameter("elliptical") || 0) * 1);
		var alpha = req.getParameter("alpha");
		var freeAR = !!((req.getParameter("freeAR") || 0) * 1);
		var alg = req.getParameter("alg") || "Smooth";
		var fmt = (req.getParameter("fmt") || "jpeg").toLowerCase();

		resp.setHeader("X-CPUCoinMiner-TweekITScriptVer", ThisScriptVersion);
		logMsg("srcDocUrl: " + srcDocUrl); // + ", w="+w+", h="+h+", x1="+x1+", y1="+y1+", x2="+x2+", y2="+y2+", alpha="+alpha+", scale="+scale+", maxPadding="+maxPadding);

		if (srcDocUrl && srcDocUrl.substr(0, 4) !== "http")
			throw "Srcdoc must use the HTTP protocol. Operation aborted.";

		// Copy the document from the dApp into the temp workdir so we always have a filename w/ an extension to use
		var ep = srcDocUrl.lastIndexOf(".");		// dApp guarantees last queried filename is the file we're interested in, so use this to get its extension
		if (ep < 0)
			throw "Srcdoc doesn't have a filename extension thus cannot be ingested: " + srcDocUrl;
		var ext = srcDocUrl.substring(ep).toLowerCase();
		params.name = workDir + "/srcdoc" + ext;
		srcDocUrl += (srcDocUrl.lastIndexOf("?") < 0 ? "?_=" : "&_=") + req.getJobId();
		docFile = new File(srcDocUrl);

		var failcount = 0;
		while (!docFile.exists() && failcount < 20)
		{
			// wait for some amount of seconds
			var startTime = new Date().getTime();
			var currTime = startTime;
			while((currTime - startTime) < 5000)
				currTime = new Date().getTime();
			failcount++;
		}
		if (failcount != 0)
			logMsg("failCount: " + failcount, LogFile.SeverityAlways);

		//logMsg("srcDocExists: " + docFile.exists(), LogFile.SeverityAlways);
		docFile.copy(params.name);
		docFile.clearCached();			// clear the downloaded original doc from the fsnet cache so we don't pollute the miner (the temp copy gets deleted automatically from the jobtemp folder when we exit)
		docFile = null;

		var xferTime = new Date().getTime();

		var image = new Media();
		var width = -1;
		var height = -1;

		if (page)
			params.page = page * 1;				// get specified page/frame
		else
			params.page = 1;					// otherwise just grab the first page
		params.frame = params.page;				// need to do this for gif and tiff src
		//logMsg("getting page " + params.frame + ", bg is " + bg);

		var mediaType = Media.getTypeFromExtension(ext);
		if (!mediaType)
			throw "Filetype '" + ext +"' is currently not a supported format.";
		if (mediaType === "MSWord" || mediaType === "MSExcel" || mediaType === "PowerPoint" || mediaType === "OODraw")
		{
			params.IntermediateFileName = params.name + ".pdf";	// load office docs as PDF first so we always get an accurate page count during preview requests. This just takes the name of the src file in the job folder and appends a proper .pdf extension.
			image.load(params);									// create the intermediate PDF
			params.name = params.IntermediateFileName;			// use this PDF as the src file now
			delete params.IntermediateFileName;					// turn this off so the load below gets an actual raster
			image.load(params);
		}
		else
			image.load(params);									// just load the raster directly from the src file

		// If a page count is requested, get it by looking at the src file or intermediate PDF. Also always get info for .png in case it is an AI image to get the generation info.
		var sdParams;
		if (getInfo || params.name.substr(-4) === ".png")
		{
			var info = Media.getFileInfo({ name: params.name });
			if (info.Frames)
				resp.setHeader("X-CPUCoinMiner-PageCount", info.Frames.toString());
			else
				resp.setHeader("X-CPUCoinMiner-PageCount", "1");
			if (info.parameters)
				sdParams = info.parameters.replace(/"/g, '&quot;');
		}

		// Save so the original image dimensions can be returned with the new ones
		var origHeight = image.getHeight();
		var origWidth = image.getWidth();

		// Process the raster
		var origPixFmt = image.getPixelFormat();
		var hasAlpha = origPixFmt.indexOf("A-") > 0;
		var forcedAlphaOn = false;
		if (fmt !== "png" && fmt !== "gif" && fmt !== "bmp")	// for anything but these 3 formats, always force the alpha output to off no matter how alpha is set by the caller
			alpha = false;
		else if (alpha == null)
			alpha = hasAlpha || elliptical;		// if no alpha param specified, default to on if output is png/gif and src has an alpha or elliptical masking is requested
		else
		{
			alpha = !!(alpha * 1);				// else if alpha param is specified and output is png/gif/bmp, use whatever the caller requested
			forcedAlphaOn = alpha;
		}

		var trueColor = true;
		if (origPixFmt.substr(0, 4) === "CMYK")
		{
			image.colorCorrect({ sourceprofile: "CMYK", destprofile: "RGB", overrideembedded: false });
		}
		else if (origPixFmt.substr(0, 3) !== "RGB")
		{
			// Make sure we're working in 24 or 32 bit for the greatest flexibility
			trueColor = false;
			image.convert({ rtype : "RGB" });
		}
		else if (image.embeddedProfile())
		{
			image.colorCorrect({ sourceprofile: "RGB", destprofile: "RGB", overrideembedded: false });	// remap any other embedded profiles to the sRGB one since now all formats support ICC or certain kinds of ICC
		}

		// Caller wants us to promote a non-alpha src image to one w/ an opaque alpha in the image data area. Padded area will be xparent.
		if (forcedAlphaOn && !hasAlpha)
			image.convert({ rtype : "RGBA" });

		// Auto-correct the orientation of the image if that metadata is present
		var rotationFixed = false;
		try
		{
			var angles = [];
			angles["6"] = 90;
			angles["3"] = 180;
			angles["8"] = 270;
			var orientation = image.getMetadata("Exif");
			if (orientation)
				orientation = orientation.match(/<Orientation>(\d+)<\/Orientation>/);
			if (orientation && orientation[1] && angles[orientation[1]])
			{
				image.rotate(angle @ angles[orientation[1]]);
				rotationFixed = true;	// don't save metadata if this is a camera image w/ rotation otherwise the image will get rotated again a viewer. TODO: just strip this one piece of metadata.
			}
			/*var exif = new ExifMetadata(false);
			exif.loadFromXML(image.getMetadata("Exif"));
			var orientation = exif.getOrientation();
			if (orientation && rotations[orientation])
				image.rotate(angle @ rotations[orientation]);
			exif.blankDocument();
			exif.setUserComment("Generated with TweekIT!");
			image.setMetadata("Exif", exif.getXML());*/
		}
		catch (e)
		{
			logMsg("Didn't get any metadata from original document because " + e, LogFile.SeverityDebug);
		}

		if (x1 < x2 && y1 < y2)
		{
			var cw = x2-x1, ch = y2-y1;
			image.crop({ xo: x1, yo: y1, xs: cw, ys: ch, relative: false, padcolor: bg })
			if (elliptical && cw > 3 && ch > 3)
			{
				var mask = new Media();
				mask.makeCanvas({ xs: cw, ys: ch, rtype: "Gray-8" });
				var xrad = Math.floor(cw / 2), yrad = Math.floor(ch / 2);
				mask.ellipse({ color: 0xFFFFFF, x: xrad, y: yrad, rx: xrad, ry: yrad, fill: true, smooth: true });
				if (alpha)
				{
					// When we can save w/ an alpha, blend the src image's alpha w/ our ellipse alpha and replace the original alpha w/ the result (don't need to worry about the color pixel data since it will be masked by the browser)
					var imask = image.clone();
					imask.exportChannel({ channel: "alpha" });					// OK if no alpha present in original - this call creates a default opaque alpha for us
					imask.composite({ source: mask, blend: "multiply" });
					image.importChannel({ channel: "alpha", source: imask });	// alpha is automatically created for non-alpha source images. Cool.
				}
				else
				{
					// Composite the bgcolor onto the image areas that are masked when there is no alpha available during the save
					image.selection({ source: mask, invert: true });
					image.composite({ color: bg });
					image.selection({ remove: true });
				}
			}
		}

		// If original had an alpha and it isn't needed replace its pixel w/ the bg
		if (hasAlpha && !alpha)
		{
			image.composite({ color: bg, blend: "Under" });
			image.convert({ rtype : "RGB" });
		}

		// Scale image to requested size, padding as needed (must be after the blend-under above due to scale creating an outline when used w/ elliptical)
		if (scale)
			image.scale({ alg: alg, xs: (scale * 100.0) + "%", constrain: true });
		else if (freeAR)
			image.scale({ alg: alg, xs: w, ys: h, constrain: false });	// caller doesn't want any constraining so the result will stretch as needed
		else
			image = scaleImage(image, w, h, bg, alg, maxPadding, cropPadding, alpha);

		// Reduce if output to GIF and alloc color for xparent if using alpha
		if (fmt === "gif")
		{
			if (alpha)
				image.reduce({ dither: trueColor ? 7 : 0, usealpha: true, alphacutoff: 128 });
			else
				image.reduce({ dither: 7 });
		}
		else if (fmt !== "jpeg" && !trueColor)
		{
			// If original image was gray or paletted and we're not saving to JPEG, convert it into a paletted img
			if (alpha)
				image.reduce({ dither: 0, usealpha: true, alphacutoff: 128 });
			else
				image.reduce({ dither: 7 });
		}

		// Pass any AI gen params through as Exif user comment. Don't need to worry about overwriting existing metadata since the AI doesn't generate Exif.
		if (sdParams)
		{
			sdParams = sdParams.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;').replace(/"/g, '&quot;').replace(/'/g, '&apos;');
			image.setMetadata('Exif', '<Exif xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"><IFD0/><IFDExif><ExifVersion>0220</ExifVersion><UserComment>' + sdParams + '</UserComment></IFDExif></Exif>');
		}

		image.setResolution({ dpi: outDpi });
		resp.setHeader("X-CPUCoinMiner-OrigImageWidth", origWidth);
		resp.setHeader("X-CPUCoinMiner-OrigImageHeight", origHeight);
		resp.setHeader("X-CPUCoinMiner-ImageWidth", image.getWidth());
		resp.setHeader("X-CPUCoinMiner-ImageHeight", image.getHeight());
		resp.setObject(image, RespType.Streamed, {type: fmt, quality: 85, jpegCompression: true, saveMetadata: !rotationFixed, embedICCProfile: fmt !== "png"});	// stream back as image w/ the specified quality if it is a JPEG (no caching) Also skip ICC due to issue w/ our sRGB and PNG.
	}
	catch (e)
	{
		if (docFile)
			docFile.clearCached();
		//resp.setStatusCode(500); // DMP: commented out status code setting because it causes the dApp web req to throw w/o being able to see the error in the header set below
		logMsg("GetImage() exception: " + e, LogFile.SeverityError);
		resp.setHeader("X-CPUCoinMiner-Error", makeAttributeValue(e.toString()));
	}

	var stopTime = new Date().getTime();
	var timing = (xferTime - startTime) + "/" + (stopTime - xferTime) + "/" + (stopTime - startTime);
	resp.setHeader("X-CPUCoinMiner-Timing", timing);
	logMsg("GetImage " + timing + " ms for xfer/processing/total of " + srcDocUrl + " to " + fmt);
}



function GetSDInfo()
{
	SetJSONResponse(new Media().aiAvailability({ detailed: true }));
}



function StableDiffuse()
{
	var startTime = new Date().getTime();

	try
	{
		var workDir = req.getJobTempDir();
		var docFile = null;
		var params = {};

		var useCLI = getPropertyValue("local.AIMediaEngine.UseCLI");
		useCLI = (!useCLI || useCLI === "false") ? false : true;
		var srcDocUrl = req.getParameter("srcdoc");
		params.dpi = (req.getParameter("dpi") || 0) * 1;
		var page = (req.getParameter("pg") || 0) * 1;
		var maxwh = (req.getParameter("maxwh") || 512) * 1;
		var x1 = (req.getParameter("x1") || 0) * 1;
		var y1 = (req.getParameter("y1") || 0) * 1;
		var x2 = (req.getParameter("x2") || 0) * 1;
		var y2 = (req.getParameter("y2") || 0) * 1;
		var w = (req.getParameter("w") || 512) * 1;
		var h = (req.getParameter("h") || 512) * 1;
		var cfgScale = (req.getParameter("cfgScale") || 7.0) * 1.0;
		var textPrompt = req.getParameter("prompt") || "nothing";
		var antiPrompt = req.getParameter("antiPrompt") || "bad anatomy, bad art, bad proportions, blurry, cloned face, cross-eye, deformed, disfigured, extra arms, extra fingers, extra legs, extra limbs, fused fingers, gross proportions, grotesque, long neck, low quality, malformed limbs, missing arms, missing legs, morbid, mutated, mutated hands, mutation, mutilated, out of frame, poorly drawn face, poorly drawn feet, poorly drawn hands, tiling, too many fingers, ugly, weird colors, worst quality";
		var denoiseStrength = req.getParameter("strength") || 0.7;
		var batchSize = (req.getParameter("batchSize") || 1) * 1;
		var steps = (req.getParameter("steps") || 20) * 1;
		var model = req.getParameter("model") || "";
		var modelHash = req.getParameter("modelHash") || "";
		var ensd = (req.getParameter("ensd") || 0) * 1;
		var seed = (req.getParameter("seed") || -1) * 1;
		var samplerName = req.getParameter("samplerName") || "";
		var upscaler = req.getParameter("upscaler") || "";
		var upscaleAmt = (req.getParameter("upscaleAmt") || 1) * 1;
		var upscalerSteps = (req.getParameter("upscalerSteps") || 0) * 1;
		var clipSkip = (req.getParameter("clipSkip") || 1) * 1;
		var lora = req.getParameter("lora") || "";
		// TODO: add iterations?

		var outputDir = workDir + (useCLI ? "" : "/samples");	// there is currently a difference in the output dir when switching between CLI and automatic1111. For automatic, match CLIs behavior of outputting into a samples subfolder

		var image = new Media();

		resp.setHeader("X-CPUCoinMiner-TweekITScriptVer", ThisScriptVersion);
		logMsg("AI Target: " + (useCLI ? "CLI" : "WebUI") + ", Batchsize: " + batchSize + ", Output to: " + outputDir + ", SamplerName: '" + samplerName + "'", LogFile.SeverityAlways);

		if (srcDocUrl)
		{
			logMsg("srcDocUrl: " + srcDocUrl + ", denoiseStrength: " + denoiseStrength + ", textPrompt: '" + textPrompt + "', antiPrompt: '" + antiPrompt + "'", LogFile.SeverityAlways);

			if (srcDocUrl && srcDocUrl.substr(0, 4) !== "http")
				throw "Srcdoc must use the HTTP protocol. Operation aborted.";

			// Copy the document from the dApp into the temp workdir so we always have a filename w/ an extension to use
			var ep = srcDocUrl.lastIndexOf(".");	// dApp guarantees last queried filename is the file we're interested in, so use this to get its extension
			if (ep < 0)
				throw "Srcdoc doesn't have a filename extension thus cannot be ingested: " + srcDocUrl;
			var ext = srcDocUrl.substring(ep).toLowerCase();
			params.name = workDir + "/srcdoc" + ext;
			srcDocUrl += (srcDocUrl.lastIndexOf("?") < 0 ? "?_=" : "&_=") + req.getJobId();
			docFile = new File(srcDocUrl);
			docFile.copy(params.name);
			docFile.clearCached();					// clear the downloaded original doc from the fsnet cache so we don't pollute the miner (the temp copy gets deleted automatically from the jobtemp folder when we exit)
			docFile = null;

			var xferTime = new Date().getTime();

			if (page)
				params.page = page * 1;				// get specified page/frame
			else
				params.page = 1;					// otherwise just grab the first page
			params.frame = params.page;				// need to do this for gif and tiff src

			var mediaType = Media.getTypeFromExtension(ext);
			if (!mediaType)
				throw "Filetype '" + ext +"' is currently not a supported format.";

			// Load the document as a raster
			image.load(params);									// just load the raster directly from the src file

			// Prepare the raster for SD
			var origPixFmt = image.getPixelFormat();
			if (origPixFmt.substr(0, 4) === "CMYK")
			{
				image.colorCorrect({ sourceprofile: "CMYK", destprofile: "RGB", overrideembedded: false });
			}
			else if (origPixFmt.substr(0, 3) !== "RGB")
			{
				// Make sure we're working in 24 or 32 bit for the greatest flexibility
				image.convert({ rtype : "RGB" });
			}
			else if (image.embeddedProfile())
			{
				image.colorCorrect({ sourceprofile: "RGB", destprofile: "RGB", overrideembedded: false });	// remap any other embedded profiles to the sRGB one since now all formats support ICC or certain kinds of ICC
			}

			// Auto-correct the orientation of the image if that metadata is present
			try
			{
				var angles = [];
				angles["6"] = 90;
				angles["3"] = 180;
				angles["8"] = 270;
				var orientation = image.getMetadata("Exif");
				if (orientation)
					orientation = orientation.match(/<Orientation>(\d+)<\/Orientation>/);
				if (orientation && orientation[1] && angles[orientation[1]])
					image.rotate(angle @ angles[orientation[1]]);
			}
			catch (e)
			{
				logMsg("Didn't get any metadata from original document because " + e, LogFile.SeverityDebug);
			}

			// Crop image sent to SD down to a user-specified focus rect.
			if (x1 < x2 && y1 < y2)
			{
				if (x1 < 0)  x1 = 0;
				if (x2 > image.getWidth())  x2 = image.getWidth();
				if (y1 < 0)  y1 = 0;
				if (y2 > image.getHeight())  y2 = image.getHeight();
				var cw = x2-x1, ch = y2-y1;
				image.crop({ xo: x1, yo: y1, xs: cw, ys: ch, relative: false })
			}

			// scale image to SD compatible size (also does SD support alpha? If not alpha will need to be stripped.)
			// Use the scaleImage function contained in this script since it already does all the cool aspect ratio preservation we need.
			// No can do - w&h need to be a multiple of 64. There is still improvement to be made here, but for now assume all submissions are going to be > 512 so just scale it to the max dims SD can handle at the moment.
			//scaleImage(image, maxwh, maxwh, 0, "Best", 100, true, false);
			var ow = image.getWidth(), oh = image.getHeight();

			// CLI version of SD requires a 512x512 src unlike automatic1111
			if (useCLI)
				image.scale({ xs: 512, ys: 512, alg: "Best" });

			// invoke img2img w/ any text prompt + the path to the image above
			logMsg("Starting img2img", LogFile.SeverityAlways);
			image.aiGenerate({ prompt: textPrompt, antiprompt: antiPrompt, seed: seed, strength: denoiseStrength, batchSize: batchSize, steps: steps, width: w, height: h, CFGScale: cfgScale, model: model, modelHash: modelHash, samplerName: samplerName, lora: lora, upscaler: upscaler, upscaleMultiplier: upscaleAmt, upscalerSteps: upscalerSteps, etaNoiseSeedDelta: ensd, lastLayerCLIPStop: clipSkip, useOutputDir: outputDir, timeout: 600 });
			logMsg("img2img finished", LogFile.SeverityAlways);
		}
		else
		{
			// No source image - use txt2img instead to generate an image from scratch
			xferTime = new Date().getTime();
			logMsg("Starting txt2img with prompt '" + textPrompt + "', antiPrompt '" + antiPrompt + "', and model '" + model + "' (" + modelHash + ")", LogFile.SeverityAlways);
			image.aiGenerate({ prompt: textPrompt, antiprompt: antiPrompt, seed: seed, strength: denoiseStrength, batchSize: batchSize, steps: steps, width: w, height: h, CFGScale: cfgScale, model: model, modelHash: modelHash, samplerName: samplerName, lora: lora, upscaler: upscaler, upscaleMultiplier: upscaleAmt, upscalerSteps: upscalerSteps, etaNoiseSeedDelta: ensd, lastLayerCLIPStop: clipSkip, useOutputDir: outputDir, timeout: 600 });
			logMsg("txt2img finished.", LogFile.SeverityAlways);
		}

		// Prepare a zip object to return all of the AI generated images
		var respPayload = new Zip();
		var samples = new File(workDir + "/samples").list();
		for (var i in samples)
		{
			var seed = samples[i].getFileNameNoExt();	// default to the existing filename (which should only be digits) if the seed extract fails
			var samplePath = samples[i].getFilePath();
			var info = Media.getFileInfo({ name: samplePath });
			if (info.parameters)
			{
				seed = info.parameters.match(/.*\r?\n.*\r?\n.+Seed: (\d+)/);
				if (seed.length > 1)
					seed = seed[1];
			}

			logMsg("Returning img " + seed + " from " + samplePath, LogFile.SeverityAlways);

			// Rescale image to original proportions, but not larger than the largest dimension returned by SD and overwrite the SD returned sample before zipping it.
			// DMP - disabled. SD is now returning a 512x512 image that has its aspect ratio reset, so rescaling it causes problems.
			if (srcDocUrl)
			{
				/*image.load({ name: samplePath });
				//image.save({ name: "temp:sdoutput"+i+".png" });
				if (ow > oh)
				{
					// Image is wider so output width is the 512. Scale height to match the original image proportions.
					image.scale({ xs: 512, ys: Math.round(oh * 512 / ow), alg: "Best" });
				}
				else
				{
					// Image is higher so output height is the 512. Scale width to match the original image proportions.
					image.scale({ ys: 512, xs: Math.round(ow * 512 / oh), alg: "Best" });
				}
				image.save({ name: samplePath });*/
			}

			respPayload.addFile(samplePath, seed + samples[i].getFileExtension());	// give the image the filename of its seed in the archive so the caller can see it w/o needing to parse the PNG
		}

		// Stream the zipped images back as the response.
		resp.setObject(respPayload, RespType.Streamed);
	}
	catch (e)
	{
		if (docFile)
			docFile.clearCached();
		logMsg("StableDiffuse() exception: " + e, LogFile.SeverityError);
		resp.setHeader("X-CPUCoinMiner-Error", makeAttributeValue(e.toString()));
	}

	var stopTime = new Date().getTime();
	var timing = (xferTime - startTime) + "/" + (stopTime - xferTime) + "/" + (stopTime - startTime);
	resp.setHeader("X-CPUCoinMiner-Timing", timing);
	logMsg("StableDiffuse " + timing + " ms for xfer/generation/total of " + srcDocUrl);
}



// Internal utility functions defined below ---------------------------------


// Scale an image, limiting to w, h, or both. Also allow limiting to amt. of padding that is visible, instead cropping the image to make it appear as such.
function scaleImage(m, targetW, targetH, bg, alg, maxPadding, cropPadding, alpha)
{
	var iw = m.getWidth();
	var ih = m.getHeight();

	// If a max ratio is set but caller only specified 1 dimension, assume the other is meant to be the same as the one they specified.
	if (maxPadding < 100)
	{
		if (!targetW && !targetH)		// if caller specified a maxPadding but didn't set any constraints, assume they want a square image limited to the image's smaller dimension
		{
			if (iw > ih)
				targetW = targetH = ih
			else
				targetW = targetH = iw;
		}
		else if (!targetW)
			targetW = targetH;
		else if (!targetH)
			targetH = targetW;
		// else both are specified so we're already all set
	}

	if (targetW && targetH)	// the fully constrained case
	{
		var didScale = false;
		var percentW = targetW / iw;	// these are the % the new dimension will be adjusted bv. If >1, we're scaling up!
		var percentH = targetH / ih;
		var verticallyLimited = percentH < percentW;
		if (maxPadding < 100)
		{
			maxPadding /= 100;				// convert percentage to rational
			var padding;
			if (verticallyLimited)
			{
				// The top and bottom of the image hit our height limit before the width, so scale to height
				var calcW = Math.round(iw * percentH);					// calc the new width based on height ratio since we're limited by that
				padding = targetW - calcW;								// this is the amt of padding that was added to the sides
				if (padding / targetW > maxPadding)						// is the image more padding on the sides than allowed?
				{
					calcW = targetW - maxPadding * targetW;				// calc what width needs to be so padding == maxPadding
					m.scale({ xs: calcW, constrain: true, alg: alg });	// scale image to the not-quite-fitting size to satisfy the maxPadding requirement
					m.crop({ xs: cropPadding ? calcW : targetW, ys: targetH, padColor: bg, transparency: alpha ? 0 : 255 });
					didScale = true;
				}
			}
			else
			{
				// The sides of the image hit our width before the height, so scale to width
				var calcH = Math.round(ih * percentW);					// calc the new height based on width ratio since we're limited by that
				padding = targetH - calcH;								// this is the amt of padding that was added to the sides
				if (padding / targetH > maxPadding)						// is the image more padding on the top and bottom than allowed?
				{
					calcH = targetH - maxPadding * targetH;				// calc what height needs to be so padding == maxPadding
					m.scale({ ys: calcH, constrain: true, alg: alg });	// scale image to the not-quite-fitting size to satisfy the maxPadding requirement
					m.crop({ xs: targetW, ys: cropPadding ? calcH : targetH, padColor: bg, transparency: alpha ? 0 : 255 });
					didScale = true;
				}
			}
		}

		// If we just needed a plain double-constrained scale, do it here
		if (!didScale)
		{
			if (cropPadding)
			{
				// Don't let scale() pad the image - calc the w & h ourselves so we only have the image itself
				if (verticallyLimited)
					m.scale({ xs: Math.round(iw * percentH), ys: targetH, constrain: false, alg: alg });
				else
					m.scale({ xs: targetW, ys: Math.round(ih * percentW), constrain: false, alg: alg });
			}
			else
			{
				m.scale({ xs: targetW, ys: targetH, constrain: true, alg: alg, padColor: bg, transparency: alpha ? 0 : 255 });
			}
		}
	}
	else if (targetW)		// if we get here, one of the two has to be unset, so check which one. This is for constrained by width
	{
		m.scale({ xs: targetW, constrain: true, alg: alg });
	}
	else if (targetH)		// and this is constrained only by height
	{
		m.scale({ ys: targetH, constrain: true, alg: alg });
	}
	// if targets are unset, don't do anything and just return image at original size

	return m;
}



function FileTypeFromExtension()
{
	var result = "";
	var ext = req.getParameter("ext");
	if (ext)
	{
		if (ext.charAt(0) != '.')
			ext = '.' + ext;
		result = Media.getTypeFromExtension(ext);
		logMsg("FileTypeFromExtension-ext: '" + ext + "', mtype: '" + result + "'");
	}
	else
	{
		var formats = Media.getFileFormats();
		for (var i=0; formats[i]; i++)
		{
			if (formats[i].flags & Media.FormatLoad)
			{
				if (result)
					result += "|";
				result += formats[i].type + ":" + formats[i].extensions;
			}
		}
	}

	SetTextResponse(result);
}



function logMsg(msg, severity) {
	var f = new LogFile("TweekIT.log");
	f.appendLine(msg, severity || LogFile.SeverityWarning);
}



function makeAttributeValue(str)
{
	var result = "";
	for (var i = 0; i < str.length; ++i)
	{
		var ch = str.charAt(i);
		var code = str.charCodeAt(i);
		result += code < 32 || (code & 0xFFFFFF80) != 0 ? "_" : ch;
	}

	return result.replace(/[^ !=#$%&*+\-.\/\^_\d\w]+/g, " ");
}
