#!/bin/bash

lpi="1.1.2.270"
text="<b>CPUCoin Miner Platform v.$lpi</b>\n\n"

getInstalled() {
	cd "$1"
	i=($(ls -1dq $2 2>/dev/null))
	if (( ${#i[@]} == 0 )); then
		echo "None"
	elif [[ $i == *.7z ]]; then
		echo "<i>Updating</i>"
	else
		echo "$i"
	fi
}

# Grab the update info and parse it into an array
vinfo=($(wget -q -O - https://dappublisher.mediarich.io/download/updates/latestversioninfoL.txt))

if [[ ${vinfo[6]} > $lpi ]]; then
	text+="This version of the CPUCoin dServices Launcher Platform is now obsolete.\n\nThe launcher platform requires an update to v.${vinfo[6]} in order to continue proper operation.\n\nPlease uninstall this version and visit us at <a href=\"https://cpucoin.io/downloads\">https://cpucoin.io/downloads</a> to download the latest platform installer."
else
	# Check to see if the miner is fully running or just in the launcher (probably updating)
	! ps -o comm= -C MediaGen 2>/dev/null | grep -x MediaGen >/dev/null 2>&1
	mining=$?
	! ps -o comm=,user= -C launch.sh 2>/dev/null | grep -x "launch.sh \+mediarich" >/dev/null 2>&1
	started=$?

	# Grab the currently installed components and place the info in vars
	cinfo=$(getInstalled /var/opt/CPUCoin/CGNMiner/clientcore/apps/client "*[0-9][0-9][0-9]*")
	winfo=$(getInstalled /var/opt/CPUCoin/CGNMiner/clientcore/apps/mws "*[0-9][0-9]*")
	text+="Installed / Available Client Build: $cinfo / ${vinfo[0]}\n\n"
	text+="Installed / Available Web Service Build: $winfo / ${vinfo[1]}\n\n"
	text+="MediaRich dService License File: ${vinfo[2]}\n\n"
	text+="Network: ${vinfo[5]}\n\n\n"
	text+="The CPUCoin Miner "
	if (( $mining )); then
		if (( $started )); then
			text+="is running.\n"
		else
			text+="is shutting down.\n"
		fi
	elif (( $started )); then
		if [[ "$cinfo" == *Updating* || "$winfo" == *Updating* ]]; then
			text+="is updating files.\n"
		else
			text+="is starting up.\n"
		fi
	else
		text+="is stopped.\n"
		if [[ "$cinfo" != ${vinfo[0]} || "$winfo" != ${vinfo[1]} ]]; then
			text+="An update will be applied upon next miner start.\n"
		fi
	fi

	text+='\n\nVisit us at <a href="https://cpucoin.io">CPUCoin.io</a>'
fi

zenity --info --no-wrap --title "About The CPUCoin dServices Miner" --window-icon /usr/lib/CPUCoin/CGNMiner/appicon.png --text "$text"
