#!/bin/bash
set +e
vinfo=()
status=252

# Check for invalid config
email=$(<"$HOME/.cgn/cgn.testnet.cpucoin.io/savedinfo/email")
entname=$(<"$HOME/.cgn/cgn.testnet.cpucoin.io/savedinfo/entityname")
if [[ ! ${email} || ! ${entname} ]]; then
	echo "Miner user information is missing. Please reinstall the app from a command prompt."
	exit 1
fi

# Download and extract the client if necessary
updateClient()
{
if [ ! -d "/var/opt/CPUCoin/CGNMiner/clientcore/apps/client/${vinfo[0]}" ]; then
	echo "Client app not present or out-of-date at /var/opt/CPUCoin/CGNMiner/clientcore/apps/client/${vinfo[0]} - updating...."
	rm -rf /var/opt/CPUCoin/CGNMiner/clientcore/apps/client/*
	wget -q -P /var/opt/CPUCoin/CGNMiner/clientcore/apps/client "https://${vinfo[3]}${vinfo[4]}/client${vinfo[0]}.7z"
	7za x -y -bd "-o/var/opt/CPUCoin/CGNMiner/clientcore/apps/client" "/var/opt/CPUCoin/CGNMiner/clientcore/apps/client/client${vinfo[0]}.7z" >/dev/null
	rm -rf "/var/opt/CPUCoin/CGNMiner/clientcore/apps/client/client${vinfo[0]}.7z"
	find /var/opt/CPUCoin/CGNMiner/clientcore/apps/client/* -type d -exec chmod 775 {} +
	find /var/opt/CPUCoin/CGNMiner/clientcore/apps/client/* -type f -exec chmod 664 {} +
	email=$(<"$HOME/.cgn/cgn.testnet.cpucoin.io/savedinfo/email")
	entname=$(<"$HOME/.cgn/cgn.testnet.cpucoin.io/savedinfo/entityname")
	minername=$(<"$HOME/.cgn/cgn.testnet.cpucoin.io/savedinfo/minername")
	sed -i -e "/MinerEmailAddress       = / s/=.*/= $email/" -e "/MinerEntityName         = / s/=.*/= $entname/" -e "/MinerDisplayName        = / s/=.*/= $minername/" -e "/MediaGenPort            = / s/=.*/= 7777/" "/var/opt/CPUCoin/CGNMiner/clientcore/apps/client/${vinfo[0]}/properties/client/MinerClient.properties"
	sed -i 's/;/:/g ; s/\\/\//g' "/var/opt/CPUCoin/CGNMiner/clientcore/apps/client/${vinfo[0]}/__start.args"
	echo -e "44,45c\nrootLogger.appenderRefs = rolling\n.\n37,42d\n11c\nappenders = rolling\n.\nw\nq" | ed -s "/var/opt/CPUCoin/CGNMiner/clientcore/apps/client/${vinfo[0]}/properties/client/log4j2.properties"
	mkdir -p "$HOME/.cgn/cgn.testnet.cpucoin.io/.credstore/miner-account"
	find "$HOME/.cgn" -type d -exec chmod 775 {} +
else
	echo "CPUCoin miner client up-to-date: ${vinfo[0]}"
fi
}

checkEOL() {
if [[ ${vinfo[6]} > "1.1.2.270" ]]; then
	echo "CPUCoin Miner Installer Platform is end-of-life (Minreq ${vinfo[6]}). A new installer needs to be downloaded and reinstalled for the miner to operate. Exiting."
	status=237
fi
}

# Download and install the latest MediaRich license if necessary
updateLicense() {
if [ ! -f "/var/opt/CPUCoin/CGNMiner/MediaRich/${vinfo[2]}" ]; then
	echo "License not present or out-of-date at /var/opt/CPUCoin/CGNMiner/MediaRich/${vinfo[2]} - updating from ${vinfo[3]}${vinfo[4]}/${vinfo[2]}."
	rm -f /var/opt/CPUCoin/CGNMiner/MediaRich/license*
	wget -q -P /var/opt/CPUCoin/CGNMiner/MediaRich "https://${vinfo[3]}${vinfo[4]}/${vinfo[2]}"
	7za x -y -bd "-o/var/opt/CPUCoin/CGNMiner/MediaRich" "/var/opt/CPUCoin/CGNMiner/MediaRich/${vinfo[2]}" >/dev/null
	cp -f /var/opt/CPUCoin/CGNMiner/MediaRich/license.bin /usr/lib/CPUCoin/CGNMiner/MediaRich/Shared/Properties/license.bin
	rm -f /var/opt/CPUCoin/CGNMiner/MediaRich/license.bin
else
	echo "CPUCoin miner license up-to-date: ${vinfo[2]}"
fi
}

# Download and extract the MediaRich dService if necessary
updateMRDS() {
if [ ! -d "/var/opt/CPUCoin/CGNMiner/clientcore/apps/mws/${vinfo[1]}" ]; then
	echo "dService endpoint not present or out-of-date at /var/opt/CPUCoin/CGNMiner/clientcore/apps/mws/${vinfo[1]} - updating...."
	rm -rf /var/opt/CPUCoin/CGNMiner/clientcore/apps/mws/*
	wget -q -P /var/opt/CPUCoin/CGNMiner/clientcore/apps/mws "https://${vinfo[3]}${vinfo[4]}/mws${vinfo[1]}.7z"
	7za x -y -bd "-o/var/opt/CPUCoin/CGNMiner/clientcore/apps/mws" "/var/opt/CPUCoin/CGNMiner/clientcore/apps/mws/mws${vinfo[1]}.7z" >/dev/null
	rm -rf "/var/opt/CPUCoin/CGNMiner/clientcore/apps/mws/mws${vinfo[1]}.7z"
	find /var/opt/CPUCoin/CGNMiner/clientcore/apps/mws/* -type d -exec chmod 775 {} +
	find /var/opt/CPUCoin/CGNMiner/clientcore/apps/mws/* -type f -exec chmod 664 {} +
	echo >>"/var/opt/CPUCoin/CGNMiner/clientcore/apps/mws/${vinfo[1]}/properties/mediarich.properties"
	sed -i -e "/SharedResultsCachePath      = / s/=.*/= \/var\/opt\/CPUCoin\/CGNMiner\/MediaRich\/Generated\/MediaResults/" -e "/SharedPropertiesPath        = / s/=.*/= \/usr\/lib\/CPUCoin\/CGNMiner\/MediaRich\/Shared\/Properties/" -e "/HttpPort                    = / s/=.*/= 7777/" "/var/opt/CPUCoin/CGNMiner/clientcore/apps/mws/${vinfo[1]}/properties/mediarich.properties"
	sed -i 's/;/:/g ; s/\\/\//g' "/var/opt/CPUCoin/CGNMiner/clientcore/apps/mws/${vinfo[1]}/__start.args"
else
	echo "dService endpoint up-to-date: ${vinfo[1]}"
fi
}

# End all miner processes. And it must be the child process - ending just the "cd 'xxxx' && ... " isn't enough.
stopMiner() {
	for pid in $(jobs -p); do
		echo "CPUCoin miner job $pid ending."
		for cpid in $(pgrep -P $pid); do
			kill -INT $cpid
		done
	done
}

# Make sure the app subfolders exist
mkdir -p /var/opt/CPUCoin/CGNMiner/clientcore/apps/client /var/opt/CPUCoin/CGNMiner/clientcore/apps/mws

# Loop until the miner is stopped or the installer is end-of-life
while (( $status == 252 )); do
	echo "Checking for CPUCoin Miner file updates...."

	# Grab the update info and parse it into an array
	vinfo=($(wget -q -O - https://dappublisher.mediarich.io/download/updates/latestversioninfoL.txt))
	echo "CPUCoin Miner version info query result: $?"

	checkEOL
	if (( $status != 237 )); then
		updateLicense
		updateMRDS
		updateClient
		stopMiner
		sleep 4

		# Launch and wait and then get the exit result from the process that ended. We only care about the client and its special needs-update status. (-4, or 252 in 8=bit bash status)
		cd "/usr/lib/CPUCoin/CGNMiner/MediaRich/Bin/MediaEngine" && ./MediaGen &
		cd "/var/opt/CPUCoin/CGNMiner/clientcore/apps/mws/${vinfo[1]}" && java -Xms128m -Dname=MediaRich -Djava.library.path=./src/main/resources/jetty -Dfile.encoding=UTF-8 @./__start.args com.equilibrium.mediarich.main.MediaRichHttpServer &
		cd "/var/opt/CPUCoin/CGNMiner/clientcore/apps/client/${vinfo[0]}" && java -Xms128m -Dname=MinerClient -Djava.library.path=./src/main/resources/bin -Dfile.encoding=UTF-8 @./__start.args io.cpucoin.miner.main.MinerClient &
		wait -n
		status=$?
		if (( $status == 252 )); then
			echo "NOTE: Miner client module has reached end-of-life. The miner will now stop, download the latest version, and restart."
		fi

	fi
done
sleep 1

stopMiner

if (( $status == 143 || $status == 137 )); then
	echo "CPUCoin Miner exit result: USER ENDED PROCESS"
elif (( $status == 237 )); then
	echo "CPUCoin Miner exit result: INSTALLER EOL"
	exit 1
elif (( $status == 0 )); then
	echo "CPUCoin Miner exit result: NORMAL EXIT"
else
	echo "CPUCoin Miner exit result: SOMETHING WENT WRONG: $status"
	exit 1
fi
